/**
 * Copyright (c) 2024 Enzien Audio, Ltd.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the phrase "powered by heavy",
 *    the heavy logo, and a hyperlink to https://enzienaudio.com, all in a visible
 *    form.
 * 
 *   2.1 If the Application is distributed in a store system (for example,
 *       the Apple "App Store" or "Google Play"), the phrase "powered by heavy"
 *       shall be included in the app description or the copyright text as well as
 *       the in the app itself. The heavy logo will shall be visible in the app
 *       itself as well.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#ifndef _HEAVY_CONTEXT_WSTD_CRSHR_HPP_
#define _HEAVY_CONTEXT_WSTD_CRSHR_HPP_

// object includes
#include "HeavyContext.hpp"
#include "HvControlCast.h"
#include "HvSignalLine.h"
#include "HvMath.h"
#include "HvControlVar.h"
#include "HvSignalVar.h"
#include "HvControlBinop.h"

class Heavy_WSTD_CRSHR : public HeavyContext {

 public:
  Heavy_WSTD_CRSHR(double sampleRate, int poolKb=10, int inQueueKb=2, int outQueueKb=0);
  ~Heavy_WSTD_CRSHR();

  const char *getName() override { return "WSTD_CRSHR"; }
  int getNumInputChannels() override { return 2; }
  int getNumOutputChannels() override { return 2; }

  int process(float **inputBuffers, float **outputBuffer, int n) override;
  int processInline(float *inputBuffers, float *outputBuffer, int n) override;
  int processInlineInterleaved(float *inputBuffers, float *outputBuffer, int n) override;

  int getParameterInfo(int index, HvParameterInfo *info) override;
  struct Parameter {
    struct In {
      enum ParameterIn : hv_uint32_t {
        CRSHR = 0xE2DA4DC3, // Crshr
        MIX = 0x97D46AC6, // Mix
      };
    };
  };

 private:
  HvTable *getTableForHash(hv_uint32_t tableHash) override;
  void scheduleMessageForReceiver(hv_uint32_t receiverHash, HvMessage *m) override;

  // static sendMessage functions
  static void cBinop_8WmIkksO_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_aCvcJ7V1_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_A6cmfKpS_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_QL5DVQuE_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cCast_O8PSDeed_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cCast_UlYmaOCI_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_Je3NquE4_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cMsg_wVeUHGcC_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_8Pwp9FV2_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_pESjofom_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_1FabOCnD_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cVar_TB3xrlyD_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cVar_aoInLSQQ_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cCast_hIlbDpLG_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cMsg_KqWVBvIL_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cMsg_hHvA4vMR_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cSend_M4NKqdVW_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_GfHPczod_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cMsg_YHhOtcy8_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cMsg_uFHLQoIk_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cMsg_KaFUZEtD_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cBinop_Zjnr6kqa_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cReceive_ya63Bkus_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cReceive_tnmGtQGR_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cReceive_Sh37w9rS_sendMessage(HeavyContextInterface *, int, const HvMessage *);
  static void cReceive_wonIjAsL_sendMessage(HeavyContextInterface *, int, const HvMessage *);

  // objects
  SignalLine sLine_wLRXE24r;
  SignalLine sLine_geDRKkNs;
  SignalLine sLine_eX5Dwy1z;
  ControlBinop cBinop_8WmIkksO;
  ControlBinop cBinop_aCvcJ7V1;
  ControlBinop cBinop_A6cmfKpS;
  ControlBinop cBinop_QL5DVQuE;
  ControlBinop cBinop_Je3NquE4;
  ControlBinop cBinop_8Pwp9FV2;
  ControlBinop cBinop_pESjofom;
  ControlBinop cBinop_1FabOCnD;
  ControlVar cVar_TB3xrlyD;
  ControlVar cVar_aoInLSQQ;
  SignalVarf sVarf_flIAf5pR;
  SignalVarf sVarf_DdHdOvSo;
  SignalVarf sVarf_rQA0sFnT;
  SignalVarf sVarf_xoGjxVMV;
  ControlBinop cBinop_GfHPczod;
  ControlBinop cBinop_Zjnr6kqa;
};

#endif // _HEAVY_CONTEXT_WSTD_CRSHR_HPP_
